(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Check_TP
imports 
  Check_DPP 
begin

subsection {*Proving termination of TRSs*}

fun
  check_tp_subsumes ::
    "('tp, ('f::show, 'l::show) lab, 'v::show) tp_ops \<Rightarrow> ('f, 'l, 'v) qreltrsLL \<Rightarrow>
    'tp \<Rightarrow> shows check"
where
  "check_tp_subsumes I (nfs', q', r', rw') tp = (do {
    let nfs = tp_ops.nfs I tp;
    let q   = tp_ops.Q I tp;
    let r   = tp_ops.R I tp;
    let rw  = tp_ops.Rw I tp;
    let rb' = r' @ rw';
    check (nfs = nfs') (shows ''incompatible substitutions-in-normal-form flags'');
    check_NF_terms_subset q q' <+? (\<lambda>r. shows ''problem with innermost strategy due to term '' +@+ shows_term r);
    check_subseteq r r'        <+? (\<lambda>r. toomuch ''rule'' (shows_rule r));
    check_subseteq rw rb'      <+? (\<lambda>r. toomuch ''relative rule'' (shows_rule r));
    return ()
  }) <+? (\<lambda>s. ''termination of the problem'' +#+ shows_nl +@+ shows_tp I tp +@+ shows_nl
           +@+ ''may not be concluded from assuming termination of the problem'' +#+ shows_nl
           +@+ shows_tp I (tp_ops.mk I nfs' q' r' rw') +@+ shows_nl
           +@+ s +@+ shows_nl)"

lemma check_tp_subsumesE[elim]:
  assumes 1: "tp_spec I"
    and 2: "isOK (check_tp_subsumes I (nfs, q, r, rw) t)"
    and 3: "tp_ops.nfs I t = nfs
    \<Longrightarrow> NF_terms (set (tp_ops.Q I t)) \<subseteq> NF_terms (set q)
    \<Longrightarrow> set (tp_ops.R I t) \<subseteq> set r
    \<Longrightarrow> set (tp_ops.Rw I t) \<subseteq> set r \<union> set rw
    \<Longrightarrow> P"
  shows "P"
  by (rule 3, insert 1 2) (simp add: Let_def | blast)+


type_synonym
  ('f, 'l, 'v) snassm = "('f, 'l, 'v) qreltrsLL list \<times> ('f, 'l, 'v) qreltrsLL"

fun mk_tp_set where
  "mk_tp_set (nfs, q, r, rw) = (nfs, set q, set r, set rw)"

fun sn_tp_assm :: "('f, 'l, 'v) snassm \<Rightarrow> bool" where
  "sn_tp_assm (ts, t) = ((\<forall>t\<in>set ts. SN_qrel (mk_tp_set t))
    \<longrightarrow> SN_qrel (mk_tp_set t))"

lemma check_tp_subsumes_sound:
  assumes I: "tp_spec I"
    and ok: "isOK (check_tp_subsumes I tp' tp)"
    and fin: "SN_qrel (mk_tp_set tp')"
  shows "SN_qrel (tp_ops.qreltrs I tp)"
proof -
  obtain nfs q r rw where tp': "tp' = (nfs, q, r, rw)" by (cases tp') force+

  let ?nfs = "tp_ops.nfs I tp"

  let ?Q   = "set (tp_ops.Q I tp)"
  let ?R   = "set (tp_ops.R I tp)"
  let ?Rw  = "set (tp_ops.Rw I tp)"
  
  let ?Q'  = "set q"
  let ?R'  = "set r"
  let ?Rw' = "set rw"

  from I and ok
    have flags: "?nfs = nfs"
    and 1: "NF_terms ?Q \<subseteq> NF_terms ?Q'" "?R \<subseteq> ?R'" "?Rw \<subseteq> ?R' \<union> ?Rw'"
    unfolding tp' by blast+
  show ?thesis unfolding tp_spec.qreltrs_sound[OF I]
  proof (rule SN_qrel_mono[OF 1])
    show "SN_qrel (?nfs, ?Q', ?R', ?Rw')"
    unfolding flags
    using fin [unfolded tp'] by auto
  qed
qed

datatype ('f, 'l, 'v) trs_termination_proof =
  DP_Trans bool bool "(('f, 'l) lab, 'v) rules" "('f, 'l, 'v) dp_termination_proof"
| Rule_Removal "('f, 'l) lab redtriple_impl" "('f, 'l, 'v) trsLL" "('f, 'l, 'v)trs_termination_proof"
| String_Reversal "('f, 'l, 'v) trs_termination_proof"
| Bounds "(('f, 'l) lab, 'v) bounds_info"
| Uncurry "(('f, 'l) lab, 'v) uncurry_info" "('f, 'l, 'v) trsLL" "('f, 'l, 'v) trs_termination_proof"
| Semlab
    "(('f, 'l) lab, 'v) sl_variant" "(('f, 'l) lab, 'v) term list"
    "('f, 'l, 'v) trsLL" "('f, 'l, 'v) trs_termination_proof"
| R_is_Empty
| Fcc "(('f, 'l) lab, 'v) ctxt list" "('f, 'l, 'v) trsLL" "('f, 'l, 'v) trs_termination_proof"
| Split "('f, 'l, 'v) trsLL" "('f, 'l, 'v) trs_termination_proof" "('f, 'l, 'v) trs_termination_proof"
| Switch_Innermost "('f, 'l) lab join_info" "('f, 'l, 'v) trs_termination_proof" 
| Drop_Equality "('f, 'l, 'v) trs_termination_proof"
| Remove_Nonapplicable_Rules "('f, 'l, 'v) trsLL" "('f, 'l, 'v) trs_termination_proof"
| Assume_SN "('f, 'l, 'v) qreltrsLL" "(('f, 'l, 'v) qreltrsLL \<times> ('f, 'l, 'v) trs_termination_proof) list"

fun assms :: "bool \<Rightarrow> ('f, 'l, 'v) trs_termination_proof \<Rightarrow> (('f, 'l, 'v) snassm + ('f, 'l, 'v) fassm) list" where
  "assms a (DP_Trans _ _ _ p) = map Inr (dp_assms a p)"
| "assms a (Rule_Removal _ _ p) = assms a p"
| "assms a (String_Reversal p) = assms a p"
| "assms a (Bounds _) = []"
| "assms a (Uncurry _ _ p) = assms a p"
| "assms a (Semlab _ _ _ p) = assms a p"
| "assms a R_is_Empty = []"
| "assms a (Fcc _ _ p) = assms a p"
| "assms a (Split _ p q) = assms a p @ assms a q"
| "assms a (Switch_Innermost _ p) = assms a p"
| "assms a (Drop_Equality p) = assms a p"
| "assms a (Remove_Nonapplicable_Rules _ p) = assms a p"
| "assms a (Assume_SN t tps) = (if a then Inl (map fst tps, t) # concat (map (assms a \<circ> snd) tps) else [])"

lemma assms_False [simp]: "assms False prf = []"
  by (induct False "prf" rule: assms.induct) simp_all

fun holds :: "('f, 'l, 'v) snassm + ('f, 'l, 'v) fassm \<Rightarrow> bool" where
  "holds (Inl p) = sn_tp_assm p"
| "holds (Inr p) = finite_dpp_assm p"

fun mk_tp where
  "mk_tp I (nfs, q, r, rw) = tp_ops.mk I nfs q r rw"

fun
  check_trs_termination_proof ::
    "bool \<Rightarrow> shows \<Rightarrow>
     ('tp, ('f::{show,key,countable}, label_type) lab, string) tp_ops \<Rightarrow>
     ('dpp, ('f, label_type) lab, string) dpp_ops \<Rightarrow> 'tp \<Rightarrow>
     ('f, label_type, string) trs_termination_proof \<Rightarrow>
     shows check" 
where
  "check_trs_termination_proof a i I J tp R_is_Empty = debug (i []) ''R is empty''
     (if tp_ops.R I tp = []
          then return () 
          else error (i +@+ '': R is not empty in the following termination-problem '' +#+
         shows_nl +@+ shows_tp I tp))"
|  "check_trs_termination_proof a i I J tp (Rule_Removal redp rR prf) = debug (i []) ''Rule Removal'' (do {
         let R = tp_ops.rules I tp;
         let Rr = ceta_list_diff R rR;
         tp' \<leftarrow> rule_removal_tt I (get_redtriple redp) Rr tp 
           <+? (\<lambda>s. i +@+ '': error when applying the rule removal technique on ''
             +#+ shows_nl +@+ shows_tp I tp +@+ shows_nl +@+
             '' to remove the rules '' +#+ shows_nl +@+ shows_rules Rr +@+ shows_nl +@+ s);
         check_trs_termination_proof a (i +@+ shows_string ''.1'') I J tp' prf
           <+? (\<lambda>s. i +@+ '': error below the rule removal technique'' +#+ shows_nl
             +@+ indent s)
   })" 
| "check_trs_termination_proof a i I J tp (DP_Trans nfs m P prf) = debug (i []) ''DP trans'' (do {
     dpp \<leftarrow> dependency_pairs_tt I J tp nfs m Sharp P
       <+? (\<lambda>s. i +@+ '': error when switching from the TRS'' +#+ shows_nl +@+ shows_tp I tp
         +@+ shows_nl +@+ ''to the initial DP problem with pairs '' +#+ shows_nl +@+ shows_rules P +@+ shows_nl +@+ s);
     check_dp_termination_proof_with_unlab a (i +@+ shows_string ''.1'') J dpp prf
       <+? (\<lambda>s. i +@+ '': error below switch to dependency pairs'' +#+ shows_nl +@+ indent s)
   })" 
|  "check_trs_termination_proof a i I J tp (String_Reversal prf) = debug (i []) ''String Reversal'' (do {
         tp' \<leftarrow> string_reversal_tt I tp 
           <+? (\<lambda>s. i +@+ '': error when applying string reversal on ''
             +#+ shows_nl +@+ shows_tp I tp +@+ shows_nl +@+ s);
         check_trs_termination_proof a (i +@+ shows_string ''.1'') I J tp' prf
           <+? (\<lambda>s. i +@+ '': error below the string reversal technique'' +#+ shows_nl
             +@+ indent s)
   })" 
|  "check_trs_termination_proof a i I J tp (Semlab sli lQ lAll prf) = debug (i []) ''Semlab'' (do {
         tp' \<leftarrow> semlab_fin_tt I sli lQ lAll tp 
           <+? (\<lambda>s. i +@+ '': error when applying semantic labelling on ''
             +#+ shows_nl +@+ shows_tp I tp +@+ shows_nl +@+ s);
         check_trs_termination_proof a (i +@+ shows_string ''.1'') I J tp' prf
           <+? (\<lambda>s. i +@+ '': error below the sem.lab technique'' +#+ shows_nl
             +@+ indent s)
   })" 
| "check_trs_termination_proof a i I J tp (Bounds info) = debug (i []) ''Bounds'' (do {
     bounds_tt I info tp 
       <+? (\<lambda>s. i +@+ '': error when applying bounds on the termination problem '' +#+ shows_nl +@+ shows_tp I tp +@+ shows_nl +@+ s)})" 
| "check_trs_termination_proof a i I J tp (Uncurry u_info rR prf) = debug (i []) ''Uncurry'' (do {
     tp' \<leftarrow> uncurry_tt I u_info rR tp
       <+? (\<lambda>s. i +@+ '': error when applying currying transformation'' +#+ shows_nl +@+ s);
     check_trs_termination_proof a (i +@+ shows_string ''.1'') I J tp' prf
       <+? (\<lambda>s. i +@+ '': error below the uncurrying technique'' +#+ shows_nl +@+ indent s)
   })"
| "check_trs_termination_proof a i I J tp (Fcc fcs R' prf) = debug (i []) ''Fcc'' (do {
    tp' \<leftarrow> fcc_tt I fcs R' tp
      <+? (\<lambda>s. i +@+ '': error when applying flat context closure'' +#+ shows_nl +@+ s);
    check_trs_termination_proof a (i +@+ shows_string ''.1'') I J tp' prf
      <+? (\<lambda>s. i +@+ '': error below flat context closure'' +#+ shows_nl +@+ indent s)
  })"
| "check_trs_termination_proof a i I J tp (Split Rrem prf1 prf2) = debug (i []) ''Split'' (do {
    let (tp1,tp2) = split_tt I tp Rrem;
    check_trs_termination_proof a (i +@+ shows_string ''.1'') I J tp1 prf1
      <+? (\<lambda>s. i +@+ '': error below the splitting'' +#+ shows_nl
        +@+ indent s);
    check_trs_termination_proof a (i +@+ shows_string ''.2'') I J tp2 prf2
      <+? (\<lambda>s. i +@+ '': error below the splitting'' +#+ shows_nl
        +@+ indent s)
  })"
|  "check_trs_termination_proof a i I J tp (Switch_Innermost joins prf) = debug (i []) ''Switch Innermost'' (do {
         tp' \<leftarrow> switch_innermost_tt I joins tp 
           <+? (\<lambda>s. i +@+ '': error when switching to innermost on ''
             +#+ shows_nl +@+ shows_tp I tp +@+ shows_nl +@+ s);
         check_trs_termination_proof a (i +@+ shows_string ''.1'') I J tp' prf
           <+? (\<lambda>s. i +@+ '': error below the switch to innermost'' +#+ shows_nl
             +@+ indent s)
   })" 
|  "check_trs_termination_proof a i I J tp (Drop_Equality prf) = debug (i []) ''Drop Equality'' (do {
         let tp' = tp_ops.mk I (tp_ops.nfs I tp) (tp_ops.Q I tp) (tp_ops.R I tp) (filter (\<lambda> (l,r). l \<noteq> r) (tp_ops.Rw I tp));
         check_trs_termination_proof a (i +@+ shows_string ''.1'') I J tp' prf
           <+? (\<lambda>s. i +@+ '': error below dropping equality rules'' +#+ shows_nl
             +@+ indent s)
   })" 
| "check_trs_termination_proof a i I J tp (Remove_Nonapplicable_Rules r prf) = debug (i []) ''Removing non-applicable rules'' (do {
       let R = tp_ops.R I tp;
       check_non_applicable_rules (tp_ops.is_QNF I tp) r
         <+? (\<lambda> s. i +@+ '': error when removing non-applicable rules'' +#+ shows_nl +@+ indent (shows_rule s +@+ shows '' is applicable''));
       let tp' = tp_spec.delete_rules I tp r;
       check_trs_termination_proof a (i +@+ shows_string ''.1'') I J tp' prf
         <+? (\<lambda>s. i +@+ '': error below the non-applicable rules removal'' +#+ shows_nl
           +@+ indent s)})"
| "check_trs_termination_proof a i I J tp (Assume_SN tp' tps) =
    debug (i []) ''Termination Assumption'' (
      if a
        then (do {
          check_tp_subsumes I tp' tp
            <+? (\<lambda>s. i +@+ '': error in termination assumption'' +#+ shows_nl +@+ s +@+ shows_nl);
          check_allm_index (\<lambda>(tp', p) j. check_trs_termination_proof a (i +@+ ''.'' +#+ shows (Suc j)) I J (mk_tp I tp') p) tps
            <+? (\<lambda>s. i +@+ '': error below termination assumption'' +#+ shows_nl
                 +@+ indent s)
        }) else error (i +@+ '': the proof contains a local termination assumption'' +#+ shows_nl)
    )"

lemma check_trs_termination_proof_with_assms_sound:
  assumes I: "tp_spec I" and J: "dpp_spec J"
    and fin: "\<forall>p\<in>set (assms a prf). holds p" (is "?P a prf")
    and ok: "isOK (check_trs_termination_proof a i I J tp prf)"
  shows "SN_qrel (tp_ops.qreltrs I tp)"
  using ok I J fin
proof (induct a i I J tp "prf" rule: check_trs_termination_proof.induct)
  case (1 a i I J tp)
  note IH = this
  interpret tp_spec I by fact
  from IH show ?case by (simp add: SN_qrel_def SN_rel_on_def)
next
  case (2 a i I J tp rp rR prof)
  interpret tp_spec I by fact
  let ?i = "i +@+ shows_string ''.1''"
  let ?Rr = "ceta_list_diff (tp_ops.rules I tp) rR"
  from 2(2) obtain tp' where
    tt: "rule_removal_tt I (get_redtriple rp) ?Rr tp = return tp'"
    and rec: "isOK (check_trs_termination_proof a ?i I J tp' prof)" by (auto simp: Let_def)
  from 2 have fin: "?P a prof" by simp
  show ?case
    apply (rule sound_tt_impl [OF rule_removal_tt [OF get_redtriple], OF tt])
    apply (rule 2(1) [OF refl refl _ rec `tp_spec I` `dpp_spec J` fin])
    apply (unfold update_error_return)
    apply (rule tt)
  done
next
  case (3 a i I J tp nfs m P prof)
  let ?i = "i +@+ shows_string ''.1''"
  from 3 obtain dpp shp where
      tt: "dependency_pairs_tt I J tp nfs m shp P = return dpp" 
      and rec: "isOK (check_dp_termination_proof_with_unlab a ?i J dpp prof)" by auto
  from 3 have fin: "\<forall>p\<in>set (dp_assms a prof). finite_dpp_assm p" by simp
  show ?case
    apply (rule dependency_pairs_tt [OF `tp_spec I` `dpp_spec J` tt])
    apply (rule check_dp_termination_proof_with_unlab_dp_assms [OF `dpp_spec J` fin rec])
  done
next
  case (4 a i I J tp prof)
  interpret tp_spec I by fact
  let ?i = "i +@+ shows_string ''.1''"
  from 4 obtain tp' where tt: "string_reversal_tt I tp = return tp'"
      and rec: "isOK (check_trs_termination_proof a ?i I J tp' prof)"
      by (auto simp: Let_def)
  from 4 have fin: "?P a prof" by simp
  show ?case
    apply (rule sound_tt_impl [OF string_reversal_tt tt])
    apply (rule 4(1) [OF _ rec `tp_spec I` `dpp_spec J` fin])
    apply (unfold update_error_return)
    apply (rule tt)
  done
next
  case (5 a i I J tp sli lQ lAll prof)
  interpret tp_spec I by fact
  let ?i = "i +@+ shows_string ''.1''"
  from 5 obtain tp' where
    tt: "semlab_fin_tt I sli lQ lAll tp = return tp'"
    and rec: "isOK (check_trs_termination_proof a ?i I J tp' prof)"
      by (auto simp: Let_def)
   from 5 have fin: "?P a prof" by simp
   show ?case
     apply (rule sound_tt_impl)
     apply (rule semlab_fin_tt)
     apply fact
     apply (rule tt)
     apply (rule 5(1) [OF _ rec `tp_spec I` `dpp_spec J` fin])
     apply (unfold update_error_return)
     apply (rule tt)
   done
next
  case (6 a i I J tp info)
  hence tt: "isOK (bounds_tt I info tp)"
    by (auto simp: Let_def)
  show ?case
    by (rule bounds_tt [OF `tp_spec I` tt])
next
  case (7 a i I J tp mode info prof)
  interpret tp_spec I by fact
  let ?i = "i +@+ shows_string ''.1''"
  from 7 obtain tp' where
    tt: "uncurry_tt I mode info tp = return tp'"
    and rec: "isOK (check_trs_termination_proof a ?i I J tp' prof)"
    by (auto simp: Let_def)
  from 7 have fin: "?P a prof" by simp
  show ?case
    apply (rule sound_tt_impl)
    apply (rule uncurry_tt)
    apply (fact)
    apply (rule tt)
    apply (rule 7(1) [OF _ rec `tp_spec I` `dpp_spec J` fin])
    apply (unfold update_error_return)
    apply (rule tt)
  done
next
  case (8 a i I J tp fcs R' prof)
  interpret tp_spec I by fact
  let ?i = "i +@+ shows_string ''.1''"
  from 8 obtain tp' where tt: "fcc_tt I fcs R' tp = return tp'"
      and rec: "isOK (check_trs_termination_proof a ?i I J tp' prof)" by auto
  from 8 have fin: "?P a prof" by simp
  show ?case
    apply (rule sound_tt_impl)
    apply (rule fcc_tt)
    apply (rule tt)
    apply (rule 8(1) [OF _ rec `tp_spec I` `dpp_spec J` fin])
    apply (unfold update_error_return)
    apply (rule tt)
  done
next
  case (9 a i I J tp R' prof1 prof2)
  interpret tp_spec I by fact
  let ?i = "\<lambda>s. i +@+ shows_string s"
  let ?i1 = "?i ''.1''"
  let ?i2 = "?i ''.2''"
  obtain tp1 tp2 where s: "split_tt I tp R' = (tp1, tp2)" by force
  from 9 s have ok1: "isOK (check_trs_termination_proof a ?i1 I J tp1 prof1)" 
      and ok2: "isOK (check_trs_termination_proof a ?i2 I J tp2 prof2)" by auto
  from 9 have fin1: "?P a prof1" and fin2: "?P a prof2" by simp_all
  show ?case
    apply (rule split_tt)
    apply (rule s)
    apply (rule 9(1) [OF _ refl ok1 `tp_spec I` `dpp_spec J` fin1])
    apply (rule s [symmetric])
    apply (rule 9(2) [OF _ refl _ ok2 `tp_spec I` `dpp_spec J` fin2])
    apply (rule s [symmetric])
    apply (unfold update_error_return)
    using ok1
    apply auto
  done
next
  case (10 a i I J tp joins prof)
  interpret tp_spec I by fact
  let ?i = "i +@+ shows_string ''.1''"
  from 10 obtain tp' where
    tt: "switch_innermost_tt I joins tp = return tp'"
    and rec: "isOK (check_trs_termination_proof a ?i I J tp' prof)"
    by (auto simp: Let_def)
  from 10 have fin: "?P a prof" by simp
  show ?case
    apply (rule sound_tt_impl)
    apply (rule switch_innermost_tt [OF `tp_spec I`])
    apply (rule infinite_lab)
    apply (rule tt)
    apply (rule 10(1) [OF _ rec `tp_spec I` `dpp_spec J` fin])
    apply (unfold update_error_return)
    apply (rule tt)
  done
next
  case (11 a i I J tp prof)
  interpret tp_spec I by fact
  let ?Rw = "filter (\<lambda> (l,r). l \<noteq> r) (Rw tp)"
  let ?Rw' = "filter (\<lambda> (l,r). l = r) (Rw tp)"
  let ?nfs = "NFS tp"
  let ?tp = "mk ?nfs (Q tp) (R tp) ?Rw"
  let ?i = "i +@+ shows_string ''.1''"
  from 11 have fin: "?P a prof" by simp
  from 11 have ok: "isOK (check_trs_termination_proof a ?i I J ?tp prof)" by auto
  have SN: "SN_qrel (qreltrs ?tp)"
    apply (rule 11(1) [OF _ ok])
    apply simp_all
    apply fact
    apply fact
    apply (rule fin)
  done
  let ?Q = "qrstep ?nfs (set (Q tp))"
  let ?QR = "?Q (set (R tp))"
  let ?QRw = "?Q (set (Rw tp))"
  let ?QRw' = "?Q (set ?Rw)"
  note d = qreltrs_sound SN_qrel_def mk_simps split
  from SN [unfolded d] have "SN_rel ?QR ?QRw'" .
  hence SN: "SN_rel ?QR (?QRw'^=)" unfolding SN_rel_Id . 
  have id: "set (Rw tp) = set ?Rw \<union> set ?Rw'" by auto
  show "SN_qrel (qreltrs tp)" unfolding d
  proof (rule SN_rel_mono [OF subset_refl _ SN], unfold id qrstep_union, rule Un_mono [OF subset_refl])
    have "?Q (set ?Rw') \<subseteq> ?Q Id" by (rule qrstep_mono, auto) 
    also have "... \<subseteq> Id" by (rule qrstep_Id)
    finally show "?Q (set ?Rw') \<subseteq> Id" .
  qed
next
  case (12 a i I J tp Rdelete prof)
  interpret tp_spec I by fact
  let ?i = "i +@+ shows_string ''.1''"
  let ?R = "delete_rules tp Rdelete"
  from 12 have ok: "isOK (check_non_applicable_rules (\<lambda>t. t \<in> NF_terms (set (Q tp))) Rdelete)"
    and rec: "isOK (check_trs_termination_proof a ?i I J ?R prof)" by auto
  from 12 have fin: "?P a prof" by simp
  from check_non_applicable_rules [OF refl ok]
    have id: "\<And>R. qrstep (NFS tp) (set (Q tp)) (R - set Rdelete) = qrstep (NFS tp) (set (Q tp)) R" .
  have "SN_qrel (tp_ops.qreltrs I
   (tp_ops.delete_R_Rw I tp Rdelete
     Rdelete))"
    apply (rule 12(1) [OF _ _ refl rec `tp_spec I` `dpp_spec J` fin])
    prefer 2
    using ok by auto
  then show ?case
    by (simp add: id SN_qrel_def)
next
  case (13 a i I J tp tp' tps)
  interpret tp_spec I by fact
  show ?case
  proof (cases a)
    assume False: "\<not> a"
    show ?thesis using False and 13 by simp
  next
    assume True: "a"
    have ok: "isOK (check_trs_termination_proof a i I J tp (Assume_SN tp' tps))" by fact
    then have subsumes: "isOK (check_tp_subsumes I tp' tp)" by simp
    have "(try check_tp_subsumes I tp' tp catch (\<lambda>x. Inl (i +@+
      '': error in termination assumption'' +#+ shows_nl +@+ x +@+
      shows_nl))) = Inr ()" using subsumes by auto
    note IH = 13(1) [OF True this _ _ _ `tp_spec I` `dpp_spec J`]
    show ?thesis
    proof (rule check_tp_subsumes_sound [OF `tp_spec I` subsumes])
      have fin: "?P a (Assume_SN tp' tps)" by fact
      then have *: "(\<And>t. t \<in> set (map fst tps) \<Longrightarrow> SN_qrel (mk_tp_set t))
        \<Longrightarrow> SN_qrel (mk_tp_set tp')" by (simp add: `a`)
      show "SN_qrel (mk_tp_set tp')"
      proof (rule *)
        fix t assume "t \<in> set (map fst tps)"
        then obtain p where "(t, p) \<in> set tps" by auto
        then obtain j where j: "j < length tps" "tps ! j = (t, p)" by (induct tps) force+
        have "SN_qrel (tp_ops.qreltrs I (mk_tp I t))"
        proof (rule IH)
          show "(t, p) \<in> set tps" by fact
        next
          show "(t, p) = (t, p)" by (rule refl)
        next
          from j and ok
            show "isOK (check_trs_termination_proof a (i +@+ ''.'' +#+ shows (Suc j)) I J (mk_tp I t) p)"
            by auto
        next
          from fin show "\<forall>a\<in>set (assms a p). holds a"
            using `(t, p) \<in> set tps` and `a` by auto
        qed
        then show "SN_qrel (mk_tp_set t)" by (cases t) simp
      qed
    qed
  qed
qed

lemma check_trs_termination_proof_sound:
  assumes "tp_spec I" and "dpp_spec J"
    and "isOK (check_trs_termination_proof False i I J tp prf)"
  shows "SN_qrel (tp_ops.qreltrs I tp)"
  using check_trs_termination_proof_with_assms_sound[OF assms(1,2) _ assms(3)] by simp

end